// -*- C++ -*-
//
// myana.cc
// main function for user analysis code
//
// Author:  Mitsuhiro YAMAGA
// Created: Thu Aug  8 23:09:32 JST 2002
//
// $Id$
//
// Revision History
//
// $Log$
//

#include <iostream>
#include <list>
#include <signal.h>
#include <string>
#include <stdlib.h>
#include "MidasManager/MidasBank.h"
#include "MidasManager/MidasFileIO.h"
#include "E391Detector/E391Detector.h"
#include "TupleManager/TupleManager.h"


// globals
MidasBank      mbank;
TupleManager   otuple;
E391Detector&  e391det = E391Detector::instance();

#define HLIM  10000000
int  pawc_[ HLIM ];


//////// user code is defined in the user_ana.cc
extern void  user_init();
extern void  user_begin_of_run( int runnum );
extern void  user_process_event( int runnum, int spillnum, int eventnum );
extern void  user_end_of_run( int runnum );
extern void  user_term();


////////
// signal handler for SIGINT
bool  Interrupted = false;
void
sighdl( int arg )
{
  Interrupted = true;
  std::cerr << "SIGINT received." << std::endl;
}



////////
// main
int
main( int argc, char* argv[] )
{
  if ( argc < 2 ){
    std::cout << "usage : " << argv[0] << " [options] filenames..."
              << std::endl 
              << "  options:" << std::endl
              << "   -o output_filename" << std::endl 
              << "   -n #event" << std::endl 
              << "   -skip skipevent" << std::endl;

    return( 1 );
  }

  // command parser
  int  Skipevent = 0;
  int  Nevent = -1;
  std::list<std::string>  flist;
  std::string  ofname;
  for( int i=1;i<argc;i++ ){
    if ( argv[i][0] == '-' ){
      if ( strcmp( argv[i],"-n" ) == 0 ){
        Nevent = strtol( argv[++i],NULL,0 );
      }
      else if ( strcmp( argv[i],"-skip" ) == 0 ){
        Skipevent = strtol( argv[++i],NULL,0 );
      }
      else if ( strcmp( argv[i],"-o" ) == 0 ){
        ofname = argv[++i];
      }
    }
    else{
      flist.push_back( std::string( argv[i] ) );
    }
  }

  // init ZEBRA
  TupleManager::init( HLIM );

  // open output file
  if ( !ofname.empty() )
    otuple.wopen( ofname,"OUT",99 );

  // call user init routine
  user_init();

  // Signal handler for SIGINT
  signal( SIGINT,sighdl );

  // main loop
  for( std::list<std::string>::iterator f = flist.begin();
       f != flist.end(); f++ ){
    // open data file
    MidasFileIO  mfile( *f,O_RDONLY|O_LARGEFILE );
    if ( mfile.good() ){
      mfile.bank( mbank );
      std::cout << "opening " << *f << std::endl;

      // event bank
      EVENT_HEADER*  pheader = mbank.eventHeader();
      void*          pevent  = mbank.eventBuffer();

      // Process Begin_of_run record
      int  run_number = -1;
      if ( mfile.readBOR() ){
        run_number = pheader->serial_number;
        std::cout << "Run=" << run_number << std::endl;
        e391det.setRunNumber( run_number );
        // get DB for this run
        //e391det.getDB();
        e391det.getDB( 1 );
        
        // call user begin_of_run routine
        user_begin_of_run( run_number );
      }

      // Process Event
      mfile.skipEvent( 1,Skipevent );
      for( int i=0; i<Nevent || Nevent<0; i++ ){
        int  id = mfile.readEvent(1);
        if ( (id & 0x8000) == 0x8000 ){
          // Special event ( BOR,EOR )
          if ( id == EVENTID_BOR ){
            run_number = pheader->serial_number;
            std::cout << "Run=" << run_number << std::endl;
            e391det.setRunNumber( run_number );
            // get DB for this run
            //e391det.getDB();
            e391det.getDB( 1 );

            // call user begin_of_run routine
            user_begin_of_run( run_number );
          }
          else if ( pheader->event_id == EVENTID_EOR ){
            // call user end_of_run routine
            user_end_of_run( run_number );
          }
        }
        else if ( id > 0 ){
          // unpack event
          e391det.unpack( pheader, pevent );
          int  spill_number = e391det.spillNumber();
          int  event_number = e391det.eventNumber();
          // call user analysis routine
          user_process_event( run_number,spill_number,event_number );
        }
        else{
          // end of file
          break;
        }
        if ( Interrupted )
          break;
      }

      // call user end_of_run routine
      user_end_of_run( run_number );

      if ( Interrupted )
        break;
    }
    else{
      std::cout << "No such file : " << *f << std::endl;
    }

  }

  // call user term routine
  user_term();

  // close 
  if ( !ofname.empty() )
    otuple.close();
  
  // signal handler for SIGINT
  signal( SIGINT,SIG_DFL );

  //
  return( 0 );
}

