// -*- C++ -*-
//
// K2917.cc
// example code
//
// Author:  Mitsuhiro YAMAGA (yamaga@post.kek.jp)
// Created: Thu May 23 22:38:26 JST 2002
//
// $Id$
//
// Revision History
//
// $Log$
//

#include <iostream>
#include <iomanip>
#include <string>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <unistd.h>
#include <sys/vme.h>
#include "k2917/K2917.h"


/*================================================*/
/* k2917.h */
/********************************************************************
 * k2917.h   Kinetic K2917 registers                                *
 *                                                                  *
 *   24-MAR-1992 Y.Takeuchi                                         *
 *                                                                  *
 * 13-JUL-1992 Y.Takeuchi    modify K2917_ENABLE,DISABLE            *
 * 30-NOV-1992 Y.Takeuchi    modify                                 *
 *  5-MAR-1993 Y.Takeuchi    modify for unified UNIX camlib         *
 * 25-OCT-1993 Y.Takeuchi    modify for Solaris2.x                  *
 ********************************************************************/

/***** DMA *****/
#define CC_DMA_DONE        0x8000
#define CC_DMA_ERR         0x1000
#define CC_DMA_ACT         0x0800
#define CC_DOCR_INIT       0x8010
#define CC_DMA_READ        0x0080
#define CC_DMA_WRITE       0x0000
#define CC_DMA_START       0x0080
#define CC_DMA_ABORT       0x0010
#define CC_DMA_INT_ENABLE  0x0008
#define CC_DMA_RESET       0x0000  

/***** Interrupt *****/
#define CC_INTR_INIT       0x0000
#define CC_TEST_FLAG       0x0080
#define CC_FLAG_AUTO_CLEAR 0x0040
#define CC_INT_ENABLE      0x0010
#define CC_INT_AUTO_CLEAR  0x0008
#define CC_IRQ1            0x0001
#define CC_IRQ2            0x0002
#define CC_IRQ3            0x0003
#define CC_IRQ4            0x0004
#define CC_IRQ5            0x0005
#define CC_IRQ6            0x0006
#define CC_IRQ7            0x0007

/***** Main register *****/
#define CC_AMR_INIT        0x007D
#define CC_CMA_INIT        0x0000

#define CC_AD              0x0001
#define CC_INLINE          0x0060
#define CC_BLOCK           0x0020
#define CC_BIT24           0x0000
#define CC_BIT16           0x0002
#define CC_QSTOP           0x0000
#define CC_QIGNO           0x0008
#define CC_QREPE           0x0010
#define CC_QSCAN           0x0018
#define CC_HALT            0x0080

#define CC_ERR             0x8000
#define CC_ABT             0x4000
#define CC_TMO             0x2000
#define CC_RST             0x1000
#define CC_LAM             0x0200
#define CC_RDY             0x0100
#define CC_DONE            0x0080
#define CC_DMA             0x0040
#define CC_WRITE           0x0020
#define CC_NOX             0x0004
#define CC_NOQ             0x0002
#define CC_GO              0x0001


// K2917
#define CC_TIMEOUT_SINGLE       1000
#define CC_STA_SINGLE_TIMEOUT   -1
#define CREATE 0


//////
#define DEBUG 0


//////
// static members


//////
// constants
static const unsigned int  K2917_MAPSIZE = 0x0100;


//////
K2917::K2917()
  : m_debugLevel( 0 ),
    m_fd16( -1 ),
    m_baseaddr( 0 ),
    m_mapsize( 0 ),
    m_mapbase( (unsigned short*)-1 ),
    m_map(     (unsigned short*)-1 ),
    m_kreg( 0 )
{
#if DEBUG
  std::cout << "K2917::K2917()" << std::endl;
#endif
}


//////
K2917::K2917( unsigned int baseaddr, int debugLevel )
  : m_debugLevel( debugLevel ),
    m_fd16( -1 ),
    m_baseaddr( baseaddr ),
    m_mapsize( 0 ),
    m_mapbase( (unsigned short*)-1 ),
    m_map(     (unsigned short*)-1 ),
    m_kreg( 0 )
{
#if DEBUG
  std::cout << "K2917::K2917(addr)" << std::endl;
#endif

  // initialize
  if ( !init( m_baseaddr ) ){
    std::cout << "K2917::K2917(addr) : init failed" 
              << std::endl;
    exit( 1 );
  }

};


//////
K2917::~K2917()
{
#if DEBUG
  std::cout << "K2917::~K2917()" << std::endl;
#endif
  
  if ( (caddr_t)m_mapbase != (caddr_t)-1 ){
    munmap( (char*)m_mapbase,m_mapsize );
  }
  
  if ( m_fd16 >= 0 ){
    close( m_fd16 );
  }
  
}


//////
bool
K2917::init( unsigned int baseaddr )
{
#if DEBUG
  std::cerr << "K2917::init()" << std::endl;
#endif

  // open VME device file 
  m_fd16 = open( "/dev/vme16d16",O_RDWR );
  if ( m_fd16 == -1 ){
    perror( "K2917::init : open /dev/vme16d16" );
    return( false );
  }

  // base addr
  m_baseaddr = baseaddr;
  unsigned int  mbase = m_baseaddr & 0xffffe000; // pagesize = 8192 = 0x2000
  
  // mmap
  m_mapsize = K2917_MAPSIZE + m_baseaddr - mbase;
  m_mapbase = (unsigned short*)mmap( 0, m_mapsize,
                                     PROT_WRITE|PROT_READ,MAP_SHARED,
                                     m_fd16, mbase );
  m_map = (unsigned short*)((unsigned int)m_mapbase + m_baseaddr - mbase);
  if ( m_debugLevel )
    std::cout << "K2917::init : mapbase=" << m_mapbase
              << std::hex
              << " : mbase=" << mbase 
              << " : map=" << m_map 
              << " : baseaddr=" << m_baseaddr
              << " : mapsize=" << m_mapsize
              << " : diff=" << (int)m_map - (int)m_mapbase
              << std::dec
              << std::endl;
  if ( (caddr_t)m_mapbase  == (caddr_t)-1 ){
    perror( "K2917::init : mmap" );
    return( false );
  }

  // K_REG
  m_kreg = (K_REG*)m_map;
  m_kreg->csr  = CC_RST;
  m_kreg->docr = CC_DOCR_INIT;

  //
  return( true );
}


//////
void
K2917::camac_init()
{
  unsigned int  dat;

  /*C*/
  camac(NAF(30, 0, 1), &dat);
  dat |= 1;
  camac(NAF(30, 0, 17), &dat);
  /*Z*/
  camac(NAF(30, 0, 1), &dat);
  dat |= 2;
  camac(NAF(30, 0, 17), &dat);
  /*SETI*/
  camac(NAF(30, 0, 1), &dat);
  dat |= 4;
  camac(NAF(30, 0, 17), &dat);
  /*REMI*/
  camac(NAF(30, 0, 1), &dat);
  dat &= ~4;
  camac(NAF(30, 0, 17), &dat);
  
}


//////
int
K2917::camac( unsigned short naf, unsigned int* dat )
{
  u_short  mode = CC_BIT24;
  int  counter,status;

  m_kreg->lamc = (u_short)CC_INT_AUTO_CLEAR;
  m_kreg->donc = (u_short)CC_INT_AUTO_CLEAR;
  m_kreg->empc = (u_short)CC_INT_AUTO_CLEAR;
  m_kreg->aboc = (u_short)CC_INT_AUTO_CLEAR;
  m_kreg->cma = CC_CMA_INIT;
  m_kreg->cmr = mode | (CREATE << 8); /* Write command list */
  m_kreg->cmr = naf;
  m_kreg->cmr = CC_HALT;
  m_kreg->cma = CC_CMA_INIT;                    /* Reset memory pointer */
  
  counter = 0;
  switch (naf & 0x0018) {
  case 0x0000:                        /* CAMAC read */
    m_kreg->lamc = (u_short)CC_INT_AUTO_CLEAR;
    m_kreg->donc = (u_short)CC_INT_AUTO_CLEAR;
    m_kreg->empc = (u_short)CC_INT_AUTO_CLEAR;
    m_kreg->aboc = (u_short)CC_INT_AUTO_CLEAR;
    m_kreg->cma = CC_CMA_INIT;
    
    m_kreg->csr &= ~CC_WRITE;
    m_kreg->csr |= CC_GO;                     /* Go! */
  
    while ((m_kreg->csr & (CC_RDY|CC_ERR)) == 0 && counter < CC_TIMEOUT_SINGLE)
      counter++;
    if ((m_kreg->csr & CC_RDY) != 0) {
      if ((mode & CC_BIT16) == 0) {
	*dat = m_kreg->dlr;
	*dat |= ((m_kreg->dhr)&0x00FF)<<16;
      }
      else
	*dat = m_kreg->dlr;
    }
        break;
  case 0x0010:                        /* CAMAC write */
    m_kreg->lamc = (u_short)CC_INT_AUTO_CLEAR;
    m_kreg->donc = (u_short)CC_INT_AUTO_CLEAR;
    m_kreg->empc = (u_short)CC_INT_AUTO_CLEAR;
    m_kreg->aboc = (u_short)CC_INT_AUTO_CLEAR;
    m_kreg->csr |= CC_WRITE;
    m_kreg->csr |= CC_GO;                     /* Go! */
    while ((m_kreg->csr & (CC_RDY|CC_ERR)) == 0 && counter < CC_TIMEOUT_SINGLE)
      counter++;
    if ((m_kreg->csr & CC_RDY) != 0) {
      if ((mode & CC_BIT16) == 0) {
	m_kreg->dlr = (*dat)&0xFFFF;
	m_kreg->dhr = (*dat)>>16;
      }
      else {
	m_kreg->dlr = *dat;
      }
    }
    break;
  default:                            /* NDT */
    m_kreg->csr |= CC_GO;             /* Go! */
    break;
  }
  while ( (m_kreg->csr & (CC_DONE | CC_ERR)) == 0 && 
          counter < CC_TIMEOUT_SINGLE)
    counter++;
  
  /*  cc->camac_qx = m_kreg->csr;*/
  
  if (counter >= CC_TIMEOUT_SINGLE) {
    status = CC_STA_SINGLE_TIMEOUT;
  }

  return(status);
}


