// -*- C++ -*-
//
// MidasFileIO.cc
// FileIO class
//
// Author:  Mitsuhiro YAMAGA (yamaga@post.kek.jp)
// Created: Tue Jul 10 18:40:57 JST 2001
//
// $Id$
//
// Revision History
//
// $Log$
//

#include <iostream>
#include <iomanip>
#include "midas.h"
#include "msystem.h"
#include "MidasManager/MidasFileIO.h"
#include "MidasManager/MidasBank.h"
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>


#ifdef OS_AIX
#include "gufio.h"
#else
#define guopen  ::open
#define guclose ::close
#define guread  ::read
#define guwrite ::write
#define gulseek ::lseek
#endif

#define  DEBUG   0

#define  ID_SPECIAL    0x8000
#define  ID_SPECIAL_R  0x0080

//////
inline bool
isSpecial( int id )
{
  return( (id & ID_SPECIAL) == ID_SPECIAL );
}


inline bool
isSpecialR( int id )
{
  return( (id & ID_SPECIAL_R) == ID_SPECIAL_R );
}


//////
MidasFileIO::MidasFileIO()
  : m_bank( NULL ),
    m_filename( "" ),
    m_flags( 0 ),
    m_permission( 0644 ),
    m_options( 0 ),
    m_fd( -1 ),
    m_fp( NULL ),
    m_zfp( NULL ),
    m_good( false )
{
  if (DEBUG) std::cerr << "MidasFileIO::MidasFileIO()" << std::endl;
}


//////
MidasFileIO::MidasFileIO( const std::string& fname, int flags, 
                          int permission )
  : m_bank( NULL ),
    m_options( 0 ),
    m_fd( -1 ),
    m_fp( NULL ),
    m_zfp( NULL )
{
  if (DEBUG) std::cerr << "MidasFileIO::MidasFileIO(fname,flags,perm,opt)"
                       << std::endl;
  open( fname,flags,permission );
}


//////
MidasFileIO::MidasFileIO( FILE* fp, int options )
  : m_bank( NULL ),
    m_filename( "" ),
    m_flags( 0 ),
    m_permission( 0644 ),
    m_options( options ),
    m_fd( -1 ),
    m_fp( NULL ),
    m_zfp( NULL ),
    m_good( false )
{
  if (DEBUG) std::cerr << "MidasFileIO::MidasFileIO(FILE*,opt)"
                       << std::endl;

  // no need to open by myself.
  open( fp,options );
}


//////
MidasFileIO::MidasFileIO( int fd, int options )
  : m_bank( NULL ),
    m_filename( "" ),
    m_flags( 0 ),
    m_permission( 0644 ),
    m_options( options ),
    m_fd( fd ),
    m_fp( NULL ),
    m_zfp( NULL ),
    m_good( true )
{
  if (DEBUG) std::cerr << "MidasFileIO::MidasFileIO(fd,opt)"
                       << std::endl;

  // no need to open by myself.
}


//////
MidasFileIO::MidasFileIO( const MidasFileIO& mio )
  : m_bank( NULL ),
    m_fd( -1 ),
    m_fp( NULL ),
    m_zfp( NULL )
{
  if (DEBUG) std::cerr << "MidasFileIO::MidasFileIO copy" << std::endl;
  open( mio.m_filename,mio.m_flags,mio.m_permission );
}


//////
MidasFileIO::~MidasFileIO()
{
  if (DEBUG) std::cerr << "MidasFileIO::~MidasFileIO" << std::endl;
  
  close();
}


//////
MidasBank*
MidasFileIO::bank( MidasBank& bank )
{
  if (DEBUG) std::cerr << "MidasFileIO::bank" << std::endl;

  MidasBank*  mb = m_bank;
  m_bank = &bank;
  return( mb );
}


//////
MidasFileIO&
MidasFileIO::operator=( const MidasFileIO& mio )
{
  if (DEBUG) std::cerr << "MidasFileIO::operator=" << std::endl;
  open( mio.m_filename,mio.m_flags,mio.m_permission );
  return( *this );
}


//////
int
MidasFileIO::open( const std::string& fname, int flags, 
                   int permission )
{
  if (DEBUG) 
    std::cerr << "MidasFileIO::open(fname,flags,perm,opt)" << std::endl;

  //
  close();

  //
  m_filename   = fname;
  m_flags      = flags;
  m_permission = permission;
  m_options    = 0;
  m_fd         = -1;
  m_fp         = NULL;
  m_zfp        = NULL;
  m_good       = false;
  
  // options
  if ( strncmp( fname.c_str(),"/hpss:",6 ) == 0 ){
    m_options |= MIO_HPSS;
  }
  if ( strncmp( fname.c_str()+fname.size()-3,".gz",3 ) == 0 ){
    m_options |= MIO_COMPRESS;
  }

  std::string  modestr;
  //if ( m_flags & O_RDONLY ) // O_RDONLY=0
  modestr = "r";
  if ( (m_flags & O_WRONLY) || (m_flags & O_CREAT) || (m_flags & O_TRUNC) )
    modestr = "w";
  if ( m_flags & O_APPEND )
    modestr = "a";
  if ( m_flags & O_RDWR )
    modestr += "+";

  //
  if ( m_options & MIO_HPSS ){
    m_fd = guopen( m_filename.c_str(),m_flags,m_permission );
    if ( m_fd >= 0 ){
      m_good = true;
      if ( DEBUG )
        std::cerr << "MidasFileIO::open : opening HPSS file : " 
                  << m_filename
                  << std::endl;
      return( 0 ); // guopen often returns errno=2 
    }
  }
  else if ( m_options & MIO_COMPRESS ){
    m_fd  = ::open( m_filename.c_str(), m_flags, m_permission );
    m_zfp = gzdopen( m_fd, modestr.c_str() );
    if ( DEBUG )
      std::cerr << "MidasFileIO::open : opening gzip file : " 
                << m_filename
                << std::endl;
    if ( m_zfp )
      m_good = true;
    else{
      ::close( m_fd );
      m_fd = -1;
    }
  }
  else{
    m_fd = ::open( m_filename.c_str(), m_flags, m_permission );
    m_fp = ::fdopen( m_fd, modestr.c_str() );
    if ( m_fp )
      m_good = true;
    else{
      ::close( m_fd );
      m_fd = -1;
    }
  }
  return( errno );
}


//////
int
MidasFileIO::open( FILE* fp, int options )
{
  if (DEBUG) std::cerr << "MidasFileIO::open(FILE*)" << std::endl;

  //
  close();

  m_bank       = NULL;
  m_filename   = "";
  m_flags      = 0;
  m_permission = 0644;
  m_options    = options;
  m_fd         = -1;
  m_fp         = NULL;
  m_zfp        = NULL;
  m_good       = true;

  // no need to open by myself.
  if ( options & MIO_COMPRESS )
    m_zfp = (void*)fp;
  else
    m_fp = fp;

  return( 0 );
}


//////
int
MidasFileIO::open( int fd, int options )
{
  if (DEBUG) std::cerr << "MidasFileIO::open(fd)" << std::endl;
 
  //
  close();

  m_bank       = NULL;
  m_filename   = "";
  m_flags      = 0;
  m_permission = 0644;
  m_options    = options;
  m_fd         = fd;
  m_fp         = NULL;
  m_zfp        = NULL;
  m_good       = true;

  return( 0 );
}



//////
int
MidasFileIO::close()
{
  if (DEBUG) std::cerr << "MidasFileIO::close" << std::endl;

  if ( m_good && !m_filename.empty() ){
    int  ret = 0;
    if ( m_zfp ){
      ret = gzclose( m_zfp );
      m_zfp = 0;
      m_fd  = -1;
    }
    else if ( m_fp ){
      ret = fclose( m_fp );
      m_fp = 0;
      m_fd = -1;
    }
    else if ( m_fd >= 0 ){
      ret = guclose( m_fd );
      if ( ret == 0 )
        m_fd = -1;
    }
  }

  m_good = false;

  return( 0 );
}


//////
int
MidasFileIO::skipEvent( int event_id,int n )
{
  // return = event_ID
  if (DEBUG) std::cerr << "MidasFileIO::skipEvent" << std::endl;

  if ( !m_good || !m_bank )
    return( 0 );
  
  int  id = 0;
  while( n > 0 ){
    if (DEBUG)
      std::cerr << "skipEvent : n=" << n << std::endl;

    id = readEvent( event_id );
    if ( isSpecial(id) || id == 0 )
      break;
    if ( id == event_id || event_id == 0 )
      n--;

    if (DEBUG) 
      std::cerr << "skipEvent : id=" << id << " event_id="<< event_id
                << std::endl;
  }
  return( id );
}

//////
int
MidasFileIO::seekEvent( int event_id, int serial, bool Rewind )
{
  // return = event_ID
  if (DEBUG) std::cerr << "MidasFileIO::seekEvent" << std::endl;

  if ( !m_good || !m_bank )
    return( 0 );
  
  if ( Rewind ){
    rewind();
  }

  int  id = 0;
  if ( serial > 0 ){
    do{
      id = readEvent( event_id );
      if ( isSpecial(id) || id == 0 ) 
        break;
    } while( m_bank->eventHeader()->serial_number != serial );
  }
  return( id );
}


//////
void
MidasFileIO::rewind()
{
  if (DEBUG) std::cerr << "MidasFileIO::rewind" << std::endl;

  if ( !m_good || !m_bank )
    return;

  if ( m_zfp )
    gzrewind( m_zfp );
  else if ( m_fp )
    ::rewind( m_fp );
  else if ( m_fd >= 0 )
    gulseek( m_fd, 0, SEEK_SET );
}


//////
int
MidasFileIO::readBOR( bool Rewind )
{
  // return = event_ID
  if (DEBUG) std::cerr << "MidasFileIO::readBOR" << std::endl;

  if ( Rewind )
    rewind();

  int  id = 0;
  while( id = readEvent( 0 ) ){
    if ( id == EVENTID_BOR )
      return( id );
  }

  // fail
  if ( Rewind )
    rewind();
  return( 0 );
}


//////
int
MidasFileIO::readEvent( int event_id )
{
  // return = event_ID
  if (DEBUG) std::cerr << "MidasFileIO::readEvent" << std::endl;
  
  if ( !m_good || !m_bank ){
    //std::cerr << "false 0" << std::endl;
    return( 0 );
  }
  
  // check buffer size
  if ( m_bank->bufsize() < sizeof(EVENT_HEADER) ){
    if ( !m_bank->realloc( sizeof(EVENT_HEADER) ) ){
      //std::cerr << "false 1" << std::endl;
      return( 0 );
    }
  }

  // read event header
  EVENT_HEADER*  pheader;
  do{
    pheader = m_bank->eventHeader();
    BANK_HEADER*  pbh = (BANK_HEADER*)(pheader + 1);
    //
    int  n = 0;
    int  rsize = sizeof(EVENT_HEADER) + sizeof(BANK_HEADER);
    if ( m_zfp )
      n = gzread( m_zfp, pheader, rsize );
    else if ( m_fp )
      n = fread( pheader, sizeof(char), rsize, m_fp );
    else if ( m_fd >= 0 )
      n = guread( m_fd, pheader, rsize );
    
    if ( n < rsize ){
      return( 0 );
    }

    // Byte Order
    // from bk_swap()
    /* only swap if flags in high 16-bit */
    bool  byte_swap = false;
    if ( isSpecialR( pheader->event_id ) || // reverse header
         ( !isSpecial( pheader->event_id ) && (pbh->flags > 0xffff) ) ){
      if ( DEBUG )
        std::cerr << "byte swap" << std::endl;

      byte_swap = true;
      WORD_SWAP(  &pheader->event_id );
      WORD_SWAP(  &pheader->trigger_mask );
      DWORD_SWAP( &pheader->serial_number );  
      DWORD_SWAP( &pheader->time_stamp );
      DWORD_SWAP( &pheader->data_size );

      if ( DEBUG )
        std::cerr << std::hex 
                  << "pheader->event_id        = " << pheader->event_id 
                  << std::endl
                  << "pheader->trigger_mask    = " << pheader->trigger_mask 
                  << std::endl
                  << "pheader->serial_number   = " << pheader->serial_number 
                  << std::endl
                  << "pheader->tiem_stamp      = " << pheader->time_stamp
                  << std::endl
                  << "pheader->data_size       = " << pheader->data_size
                  << std::endl
                  << "pbank->data_size(not swapped) = " << pbh->data_size 
                  << std::endl
                  << "pbank->flags(not swapped)     = " << pbh->flags 
                  << std::endl;
    }

    // read event  
    if ( pheader->data_size > 0 ){
      if ( m_bank->bufsize() < pheader->data_size+sizeof(EVENT_HEADER) ){
        if ( !m_bank->realloc( pheader->data_size+sizeof(EVENT_HEADER) ) ){
          //std::cerr << "false 3" << std::endl;
          perror( "MidasFileIO::readEvent" );
          return( 0 );
        }
      }
      // read
      int  rsize = pheader->data_size - sizeof(BANK_HEADER);
      if ( m_zfp )
        n = gzread( m_zfp, pbh+1, rsize );
      else if ( m_fp )
        n = fread( pbh+1, sizeof(char), rsize, m_fp );
      else if ( m_fd >= 0 )
        n = guread( m_fd, pbh+1, rsize );
      // byte swap?
      if ( byte_swap && 
           (pheader->event_id & ID_SPECIAL)!=ID_SPECIAL ){ // normal event
        // swap the event, including BANK_HEADER
        int  ret = bk_swap( pheader+1,false );
        if ( DEBUG )
          std::cerr << "bk_swap : ret=" << ret << std::endl;
      }
      //
      if ( DEBUG )
        std::cerr << "read " << n << " bytes :"
                  << " datasize=" << pheader->data_size 
                  << std::endl;
      
      if ( n < rsize ){
        std::cerr << "MidasFileIO::readEvent : unexpected EOF" << std::endl;
        return( 0 );
      }
    }
    if ( isSpecial( pheader->event_id ) ){
      return( pheader->event_id ); // special event
    }
  } while( pheader->event_id != event_id && event_id != 0 );
  return( pheader->event_id );
}


//////
int
MidasFileIO::writeEvent()
{
  // return = size
  if (DEBUG) std::cerr << "MidasFileIO::writeEvent" << std::endl;

  if ( !m_good || !m_bank )
    return( 0 );

  //
  m_bank->eventHeader()->data_size = bk_size( m_bank->eventHeader() + 1 );
  int  dtsize = m_bank->eventHeader()->data_size + sizeof(EVENT_HEADER);
  
  //
  int  n = 0;
  if ( m_zfp )
    n = gzwrite( m_zfp, m_bank->eventHeader(), dtsize );
  else if ( m_fp )
    n = fwrite( m_bank->eventHeader(), sizeof(char), dtsize, m_fp );
  else if ( m_fd >= 0 )
    n = guwrite( m_fd, m_bank->eventHeader(), dtsize );
  
  if (DEBUG){
    std::cerr << "MidasFileIO::writeEvent : data=" 
              << std::hex
              << dtsize << " write=" << n
              << std::dec << std::endl;
  }
  
  if ( n < dtsize ){
    std::cerr << "MidasFileIO::writeEvent : cannot write data :"
              << " wrote=" << n << " datasize=" << dtsize
              << std::endl;
    return( 0 );
  }
  return( n );
}


//////
int
MidasFileIO::writeSpecialEvent()
{
  // return = size
  if (DEBUG) std::cerr << "MidasFileIO::writeSpecialEvent" << std::endl;

  if ( !m_good || !m_bank )
    return( false );

  //
  int  dtsize = m_bank->eventHeader()->data_size + sizeof(EVENT_HEADER);
  
  //
  int  n = 0;
  if ( m_zfp )
    n = gzwrite( m_zfp, m_bank->eventHeader(), dtsize );
  else if ( m_fp )
    n = fwrite( m_bank->eventHeader(), sizeof(char), dtsize, m_fp );
  else if ( m_fd >= 0 )
    n = guwrite( m_fd, m_bank->eventHeader(), dtsize );
  
  if (DEBUG){
    std::cerr << "MidasFileIO::writeSpecialEvent : data=" 
              << std::hex
              << dtsize << " write=" << n
              << std::dec << std::endl;
  }

  if ( n < dtsize ){
    std::cerr << "MidasFileIO::writeEvent : cannot write data :"
              << " wrote=" << n << " datasize=" << dtsize
              << std::endl;
    return( 0 );
  }
  return( n );
}


